# $Id: sdb.py,v 1.5 2005/11/08 18:46:22 tofu Exp $
"""
 A in memory session manager for punjab.
"""
from twisted.python import components
from twisted.web import server, resource
from twisted.internet import reactor, protocol, defer
try:
    from twisted.words.xish import domish, utility
except:
    from twisted.xish import domish, utility

from twisted.application import internet, service

from zope.interface import Interface, implements

from xmpp import component

import traceback
import md5, random
import sha, base64

# this is temp
NS_SDB = 'http://butterfat.net/sdb'


class Session(server.Session):
    values = {}
    
    def __setitem__(self, name, value):
        self.values[name] = value

    def __getitem__(self, name):
        return self.values[name]

class SdbParse:
    def __init__(self):
        # Setup the parser
        self.stream = domish.elementStream()
        self.stream.DocumentStartEvent = self.onDocumentStart
        self.stream.ElementEvent = self.onElement
        self.stream.DocumentEndEvent = self.onDocumentEnd
        
        self.data = None
        

    def parse(self,buf):
        self.stream.parse(buf)
        return self.data
    
    def serialize(self, obj):
        if isinstance(obj, domish.Element):
            obj = obj.toXml()
        return obj

    def onDocumentStart(self, rootelem):
        self.data = rootelem

    def onElement(self, element):
        if isinstance(element, domish.Element):
            self.data.addChild(element)
        else:
            pass

    def _reset(self):
        # Setup the parser
        self.stream = domish.elementStream()
        self.stream.DocumentStartEvent = self.onDocumentStart
        self.stream.ElementEvent = self.onElement
        self.stream.DocumentEndEvent = self.onDocumentEnd
        self.data = None
        

    def onDocumentEnd(self):
        pass

        
class ISdbService(Interface):


    def startSession(self):
        """ Start a punjab jabber session """

    def endSession(self, session):
        """ end a punjab jabber session """

    def setValue(self, name, value):
        """ set a value in the db """

    def getValue(self, name):
        """ get a value from the db """
        
        
class SdbService:
    """
    A session service for storing name value pairs

    NOTE : some code taken from twisted.web.server.Session (THANKS)
    """
    implements(ISdbService)

    def __init__(self):
        self.counter  = 0
        self.sessions = {}
        self.clients  = 0

    def _mkuid(self):
        """(internal) Generate an opaque, unique ID for a user's session.
        """
        #print ip
        self.counter = self.counter + 1
        return md5.new("%s_%s" % (str(random.random()) , str(self.counter))).hexdigest()
        
    
    def makeSession(self, ip):
        """Generate a new Session instance, and store it for future reference.
        """
        uid = self._mkuid()
        s = Session(self, uid)
        sid = s.sid = md5.new("%s_%s" % (str(uid) , str(ip))).hexdigest()
        
        session = self.sessions[sid] = s
        
        reactor.callLater(1800, s.checkExpired)
        return session

    def getSession(self, sid, ip = None):
        """Get a previously generated session, by its unique ID.
        """
        if self.sessions.has_key(sid):
            return self.sessions[sid]
        else:
            return None
        

    def endSession(self, sid):
        if self.sessions.has_key(sid):
            del self.sessions[sid]

    def setValue(self, sid, name, value):
        """ set a value in the db """
        if self.sessions.has_key(sid):
            self.sessions[sid].values[name] = value

    def getValues(self, sid):
        """ get a value from the db """
        if self.sessions.has_key(sid):
            return self.sessions[sid].values

    def getValue(self, sid, name):
        """ get a value from the db """
        if self.sessions.has_key(sid):
            return self.sessions[sid].values[name]

    def delValue(self, sid, name):
        """ delete a value from the db """
        if self.sessions.has_key(sid):
            del self.sessions[sid].values[name]
    
class Sdb(resource.Resource):
    isLeaf = True
    def __init__(self, service, v=0):
        """Initialize.
        """
        resource.Resource.__init__(self)
        self.service  = service

    def return_request(self, request, d, charset="utf-8"):
        request.setResponseCode(200)
        if isinstance(d, domish.Element):
            xml = unicode(d.toXml()).encode(charset,'replace')
        else:
            xml = d
            
        request.setHeader('content-type', 'text/xml')
        request.setHeader('content-length', len(xml))
        request.write(xml)
        request.finish()

    def return_error(self, request, selem, error):
        selem.addElement('error', None, error)
        self.return_request(request, selem)
        
    def getChild(self, selem, request):
        # NOTE - the path with be the values you would like to retieve
        uid = selem['sid']
        sid = md5.new("%s_%s" % (str(uid) , str(request.client.host))).hexdigest()
        
        if request.postpath[0] != '':
            # get parts
            for n in request.postpath:
                val = self.service.getValue(sid,n)
                selem.addElement(n, None, val)
        else:
            # get everything
            vals = self.service.getValues(sid)
            for k in vals:
                selem.addElement(k, None, vals[k])
            
        
        self.return_request(request, selem)

        
    # TODO - Do a key based session query? Also, check for ip
    def render_GET(self, request):
                
        s = domish.Element((NS_SDB,'session'))

        if len(request.received_cookies)==0 or not request.received_cookies.has_key('sid'):
            self.return_error(request, s, 'no session')
        else:
            # make this cookie based
            s['sid'] = request.received_cookies['sid']
            self.getChild(s, request)
        return server.NOT_DONE_YET

    def render_POST(self, request):
        # make this cookie based
        request.content.seek(0, 0)       
        # create a unique session id
        s  = self.service.makeSession(request.client.host)
                
        sp = SdbParse()
        
        sc = sp.parse(request.content.read())
        
        for e in sc.elements():
            self.service.setValue(s.sid,e.name.encode('ascii','replace'), str(e))

        nsc = domish.Element((NS_SDB,'session'))
        nsc.addElement('sid',None,s.uid)
        request.headers['set-cookie'] = 'sid='+s.uid
        self.return_request(request, nsc)
        return server.NOT_DONE_YET
        


    def render_PUT(self, request):
        request.content.seek(0, 0)       
        sp = SdbParse()

        sc = sp.parse(request.content.read())
        
        if len(request.received_cookies)==0 or not request.received_cookies.has_key('sid'):
            self.return_error(request, sc, 'no session')
        else:
            # make this cookie based
            uid = request.received_cookies['sid']
            sid = md5.new("%s_%s" % (str(uid) , str(request.client.host))).hexdigest()
        
            nsc = domish.Element((NS_SDB,'session'))
            for e in sc.elements():
                self.service.setValue(sid,str(e.name), str(e))
                nsc.addChild(e)
            self.return_request(request, nsc)
                
        return server.NOT_DONE_YET
        

    def render_DELETE(self, request):
        nsc = domish.Element((NS_SDB,'session'))
        if len(request.received_cookies)==0 or not request.received_cookies.has_key('sid'):
            self.return_error(request, nsc, 'no session')
        else:
            # make this cookie based
            uid = request.received_cookies['sid']
            sid = md5.new("%s_%s" % (str(uid) , str(request.client.host))).hexdigest()
            
            ecount = 0
            for e in request.postpath:
                ecount = ecount + 1
                self.service.delValue(sid,e)
                nsc.addElement(e)
            if ecount == 0:
                self.server.endSession(sid)

            self.return_request(request, nsc)
        return server.NOT_DONE_YET
        


    def return_session(self, data, session, request):
        """
        Return a session xml element
        """
        # create session element
        
        s = domish.Element("session")       
        
        request.setResponseCode(200)
        bxml = s.toXml().encode('utf-8', 'replace')
        request.setHeader('content-type', 'text/xml')
        request.setHeader('content-length', len(bxml))
        request.write(bxml)
        request.finish()
            
    def send_http_error(self,code,request, error ='', charset = 'utf-8'):
        request.setResponseCode(code)
        
        s = domish.Element("session")
        sxml = s.toXml().encode('utf-8', 'replace')
        request.setHeader("content-length", len(sxml))    
        request.write(sxml)
        
        #request.setHeader("content-length", "0")
        request.finish()


components.registerAdapter(Sdb, ISdbService, resource.IResource)

